<?php

    include '../../../common/include/common.php';
    include_once '../../config.php';
    printHeader($module, "Tool Documentation Index", "Guide");
?>

<div class='row-fluid'>

<div class='span9'>



<h1 id="top">Tool Documentation Index
    <small>11.1</small>
</h1>
<div class="accordion" id="index">
    <div class="accordion-group">
        <div class="accordion-heading">
            <a class="accordion-toggle" data-toggle="collapse" data-parent="#index" href="#WDLfeaturegroupname">
                <h4>WDL feature group name</h4>
            </a>
        </div>
        <div class="accordion-body collapse" id="WDLfeaturegroupname">
            <div class="accordion-inner">
                <p class="lead">Test program group used for testing</p>
                <table class="table table-striped table-bordered table-condensed">
                    <tr>
                        <th>Name</th>
                        <th>Summary</th>
                    </tr>
                            <tr>
                                    <td><a href="org_broadinstitute_barclay_help_wdltestinputs_TestWDLTool.wdl">TestWDLTool</a></td>
                                <td>WDL Test Tool to test WDL Generation</td>
                            </tr>
                </table>
            </div>
        </div>
    </div>
</div>

        <hr>
        <p><a href='#top'><i class='fa fa-chevron-up'></i> Return to top</a></p>
        <hr>
        <p class="version">Barclay version 11.1 built at 2016/01/01 01:01:01.
        </p>

</div></div>

<?php

    printFooter($module);

?>