<?php
    include '../../../common/include/common.php';
?>

<div class='row-fluid' id="top">

	

	<?php $group = 'Test feature group name'; ?>

	<section class="span4">
		<aside class="well">
			<a href="index"><h4><i class='fa fa-chevron-left'></i> Back to Tool Docs Index</h4></a>
		</aside>
		<aside class="well">
			<h2>Categories</h2>
        <style>
            #sidenav .accordion-body a {
                color : gray;
            }

            .accordion-body li {
                list-style : none;
            }
        </style>
        <ul class="nav nav-pills nav-stacked" id="sidenav">
        		<hr>
        		<hr>
        		<hr>
						<li><a data-toggle="collapse" data-parent="#sidenav" href="#Testextradocsgroupname">Test extra docs group name</a>
							<div id="Testextradocsgroupname"
								<?php echo ($group == 'Test extra docs group name')? 'class="accordion-body collapse in"'.chr(62) : 'class="accordion-body collapse"'.chr(62);?>
								<ul>
											<li>
												<a href="org_broadinstitute_barclay_help_testinputs_TestExtraDocs.html">TestExtraDocs</a>
											</li>
								</ul>
							</div>
						</li>
						<li><a data-toggle="collapse" data-parent="#sidenav" href="#Testfeaturegroupname">Test feature group name</a>
							<div id="Testfeaturegroupname"
								<?php echo ($group == 'Test feature group name')? 'class="accordion-body collapse in"'.chr(62) : 'class="accordion-body collapse"'.chr(62);?>
								<ul>
											<li>
												<a href="org_broadinstitute_barclay_help_testinputs_TestArgumentContainer.html">TestArgumentContainer</a>
											</li>
											<li>
												<a href="org_broadinstitute_barclay_help_testinputs_TestDeprecatedCLP.html">TestDeprecatedCLP</a>
											</li>
								</ul>
							</div>
						</li>
        		<hr>
        </ul>
		</aside>
		<?php getForumPosts( 'TestDeprecatedCLP' ) ?>

	</section>

	<div class="span8">

		<h1>TestDeprecatedCLP **DEPRECATED**</h1>
		<h2>This feature is deprecated and will be removed in a future release.</h2>

		<p class="lead">Argument container class for testing deprecated CLPs.</p>

			<h3>Category
				<small> Test feature group name</small>
			</h3>
		<hr>
		<h2>Overview</h2>
		A test tool for testing a deprecated (@DeprecatedFeature) CLP.

				<hr>
				<h2>Command-line Arguments</h2>
				<p></p>
				<h3>Additional References</h3>
				<p>See these additional references for more information.</p>
				<ul>
						<li><a href="org_broadinstitute_barclay_help_testinputs_TestExtraDocs.html">TestExtraDocs</a></li>
				</ul>

				<h3>TestDeprecatedCLP specific arguments</h3>
				<p>This table summarizes the command-line arguments that are specific to this tool. For more details on each argument, see the list further down below the table or click on an argument name to jump directly to that entry in the list.</p>
				<table class="table table-striped table-bordered table-condensed">
					<thead>
					<tr>
						<th>Argument name(s)</th>
						<th>Default value</th>
						<th>Summary</th>
					</tr>
					</thead>
					<tbody>
			<tr>
				<th colspan="4" id="row-divider">Advanced Arguments</th>
			</tr>
				<tr>
					<td><a href="#--optionalInt">--optionalInt</a><br />
								&nbsp;<em>-optInt</em>
					</td>
					<!--<td>int</td> -->
					<td>1</td>
					<td>		
optionalInt with initial value 1</td>
				</tr>
					</tbody>
				</table>

					<h3>Argument details</h3>
					<p>Arguments in this list are specific to this tool. Keep in mind that other arguments are available that are shared with other tools (e.g. command-line GATK arguments); see Inherited arguments above.</p>
		<hr style="border-bottom: dotted 1px #C0C0C0;" />
		<h3><a name="--optionalInt">--optionalInt </a>
			 / <small>-optInt</small>
		</h3>
		<p class="args">
			<b>optionalInt with initial value 1</b><br />
			Optional int
		</p>
		<p>
			<span class="label label-info ">int</span>
				&nbsp;<span class="label">1</span>
				&nbsp;<span class="label label-warning">[ [ -∞</span>
				&nbsp;<span class="label label-warning">∞ ] ]</span>
		</p>

        <hr>
        <p><a href='#top'><i class='fa fa-chevron-up'></i> Return to top</a></p>
        <hr>
        <p class="version">Barclay version 11.1 built at 2016/01/01 01:01:01.
        </p>

	</div>

	<?php printFooter($module); ?>