package org.broadinstitute.barclay.help;

/**
 * Base class for documentation generation exceptions generated by Barclay.
 */
public class DocException extends RuntimeException {
    private static final long serialVersionUID = 1L;

    public DocException( String msg ) {
        super(msg);
    }

    public DocException( String message, Throwable throwable ) {
        super(message, throwable);
    }

}
