package org.broadinstitute.barclay.help;

/**
 * Public constants for keys used with the external (FreeMarker) template map. These strings must
 * match the strings that are embedded/referenced in the templates code.
 */
public class TemplateMapConstants {
    // GSON Arg properties
    public final static String GSON_ARG_DEPRECATED = "deprecated";
    public final static String GSON_ARG__DEPRECATION_DETAIL = "deprecationDetail";

    // Arg "extra" attributes
    public final static String ARG_DEPRECATED_ATTRIBUTE = "deprecated";

    public final static String FEATURE_DEPRECATION_DETAIL = "deprecationDetail";

    public final static String GSON_FEATURE_DEPRECATED = "deprecated";
    public final static String GSON_FEATURE_DEPRECATED_DETAIL = "deprecationDetail";

}
