package org.broadinstitute.barclay.help.testinputs;

import org.broadinstitute.barclay.argparser.Argument;

import java.io.File;
import java.util.List;

/**
 * Base class definition for an argument collection that can be subclassed and instantiated.
 */
public class TestBaseArgumentType {
}
