package org.broadinstitute.barclay.help.testinputs;

public enum TestNonCLPEnum {
    /**
     * This is the NON_CLP_ENUM_VALUE_1 comment
     */
    NON_CLP_ENUM_VALUE_1,

    /**
     * This is the NON_CLP_ENUM_VALUE_2 comment
     */
    NON_CLP_ENUM_VALUE_2
}
