package org.broadinstitute.barclay.help;

/**
 * Constants for the Barclay options.
 */
public class BarclayDocletOptions {

    // Builtin javadoc command line arguments
    final static String DESTINATION_DIR_OPTION = "-d";
    final static String WINDOW_TITLE_OPTION = "-windowtitle";
    final static String DOC_TITLE_OPTION = "-doctitle";
    final static String QUIET_OPTION = "-quiet";

    // Barclay HelpDoclet custom command line options
    final static String SETTINGS_DIR_OPTION = "-settings-dir";
    final static String BUILD_TIMESTAMP_OPTION = "-build-timestamp";
    final static String ABSOLUTE_VERSION_OPTION = "-absolute-version";
    final static String INCLUDE_HIDDEN_OPTION = "-hidden-version";
    final static String OUTPUT_FILE_EXTENSION_OPTION = "-output-file-extension";
    final static String INDEX_FILE_EXTENSION_OPTION = "-index-file-extension";
    final static String USE_DEFAULT_TEMPLATES_OPTION = "-use-default-templates";
}
